from collections import defaultdict
djangoToRestRegistry = defaultdict(None) #a dictionary which is used as a registry to store 
#all the components of the project like serializers, viewsets, actions, custom parameters, etc